/************************************************************************************
  Sensor_r01.ino   LN Version
/************************************************************************************* 
   This sketch requires the following libraries:
    Loconet Library 1.1.4 

   2022/01/22 Rev01
   Copyright (c) 2022 Nagoden  All right reserved.
*************************************************************************************/
//#include <EEPROM.h>

//#define DEBUG

// include the library code:
// https://github.com/Seeed-Studio/Grove_4Digital_Display
#include "TM1637.h"
//----------------------------------------------------------------------
// Settings used in Loconet
#include <LocoNet.h>
#define DAT_PIN0 A0         //output Sw PORT 1　検出表示LED用
//#define DAT_PIN1 A1         //output Sw PORT 2　未使用
//#define DAT_PIN2 A2         //output Sw PORT 3　未使用
//#define DAT_PIN3 A3         //output Sw PORT 4　未使用
//#define DAT_PIN4 3          //output Sw PORT 5　未使用
//#define DAT_PIN5 4          //output Sw PORT 6　未使用
//#define DAT_PIN6 5          //output Sw PORT 7　未使用
#define DAT_PIN7 6          //output Sw PORT 8　在線検出センサーインプット
#define SETUP_PIN 9         //input SetupPin
#define STAT_LED_PIN 10     // Status_LED on/off pin
//Address Disp const　      ボードの設定値は以下の３つを設定変更する
#define BOARDID 3           //BoardID
#define SECTIONID 9         //SectionID
#define ZONEID 5
//ZONE  A:1 B:3 C:5 D:7 E:9 F:11 G:13 H:15
//１つのボードアドレスでアドレス表示できるゾーンは８つ
//ゾーンとセクションの関係は下記の通り
//BoardID       Zone          Section
//1 to 255      A:1           1  
//1 to 255      A:1           2  
//1 to 255      A:1           3  
//1 to 255      A:1           4  
//1 to 255      B:3           5
//1 to 255      B:3           6
//1 to 255      B:3           7
//1 to 255      B:3           8
//1 to 255      C:5           9
//1 to 255      C:5           10
//1 to 255      C:5           11
//1 to 255      C:5           12
//1 to 255      D:7           13
//1 to 255      D:7           14
//1 to 255      D:7           15
//1 to 255      D:7           16
//1 to 255      E:9           1  
//1 to 255      E:9           2  
//1 to 255      E:9           3  
//1 to 255      E:9           4  
//1 to 255      F:11          5
//1 to 255      F:11          6
//1 to 255      F:11          7
//1 to 255      F:11          8
//1 to 255      G:13          9
//1 to 255      G:13          10
//1 to 255      G:13          11
//1 to 255      G:13          12
//1 to 255      H:15          13
//1 to 255      H:15          14
//1 to 255      H:15          15
//1 to 255      H:15          16
/*********************************************************************************
 Variable definition Address Disp
 ********************************************************************************/
uint16_t  gBoard = BOARDID;          //ボードID
uint8_t   gSection_no = SECTIONID;   //セクション
int       gZone  = ZONEID;           //ゾーン
int       gPresent = 0;              //検出スイッチ 
int       gSection_mem = 0;          //セクションフラグ
int       gAddress_mem = 0;          //アドレスフラグ

// Pins definitions for TM1637
const int CLK = A5;     //PC6
const int DIO = A4;     //PC5
// TM1637 and can be changed to other ports
TM1637 tm1637(CLK, DIO);
/*********************************************************************************
 Variable definition Address Decode
 ********************************************************************************/

#define TIME_RECVWAIT 25
#define SIZE_RECV 16
#define SIZE_CONV 8

#define CHK_RCID2 B0010 //アドレスLow
#define CHK_RCID1 B0001 //アドレスHigh

typedef struct {
    byte mID;
    byte mData1; //12bit
    byte mData2; //18bit
    byte mData3; //24bit / 12bit + 12bit
    byte mData4; //36bit
} RAILCOM_DATAGRAM;

//Codings in the 4 out of 8 code NMRA RCN-217 Table2
//変換テーブル
const int convtable[256] = {
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 51, 255, 255, 255, 52, 255, 53, 54, 255, 255, 255, 255,
    255, 255, 255, 255, 58, 255, 255, 255, 59, 255, 60, 55, 255, 255, 255, 255, 63, 255,
    61, 56, 255, 255, 62, 57, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
    36, 255, 255, 255, 35, 255, 34, 33, 255, 255, 255, 255, 31, 255, 30, 32, 255, 255, 29,
    28, 255, 27, 255, 255, 255, 255, 255, 255, 25, 255, 24, 26, 255, 255, 23, 22, 255, 21,
    255, 255, 255, 255, 37, 20, 255, 19, 255, 255, 255, 50, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 14, 255, 13, 12, 255, 255,
    255, 255, 10, 255, 9, 11, 255, 255, 8, 7, 255, 6, 255, 255, 255, 255, 255, 255, 4, 255,
    3, 5, 255, 255, 2, 1, 255, 0, 255, 255, 255, 255, 15, 16, 255, 17, 255, 255, 255, 18,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 43, 48, 255, 255, 42, 47,
    255, 49, 255, 255, 255, 255, 41, 46, 255, 45, 255, 255, 255, 44, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 40, 255, 39, 255, 255, 255, 38, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255
};

//表示アドレス
uint16_t gDetectedLocAddress;
word gAddrDataLow = 0;
word gAddrDataHigh = 0;
word gAddrReceived = 0;

//制御用変数
uint8_t interruptflag = 0;
long no_signaltime = 0;
int no_recive = 0;
long no_commandtime = 0;
int no_command = 0;
byte gWaitTime = 0; //CutOutからの受信ウェイト用(50ms)
byte gLedSegOnOff = 0; //7セグの点滅表示
byte gRailComCutOutDetect = 0;

//受信データの変数
byte receivedByte = B00000000;
int command_count = 0;
byte gReciveData[SIZE_RECV] = {
    0
};
byte gConversionData[SIZE_CONV] = {
    0
};

byte bitwork1 = B00000000;
byte bitwork2 = B00000000;
byte bitworkA = B00000000;
byte bitworkB = B00000000;
RAILCOM_DATAGRAM gRecvDataGram[SIZE_CONV];


//work コマンド通信用
byte sendwork_aa = B00000000;
byte sendwork_bb = B00000000;
byte sendwork_cc = B00000000;
byte sendwork_dd = B00000000;
byte sendchksum = B00000000;
uint8_t Low_add = 0 ;
uint8_t High_add= 0 ;

void receiveSerial();
void excec_conversion(byte inRecvSize);
void clearAddr();

/*********************************************************************************
 Loconet Class定義
 ********************************************************************************/
LocoNetClass  Loconet;
lnMsg       *LnPacket;
LnBuf       LnTxBuffer;
/*********************************************************************************
 Task Schedule
 ********************************************************************************/
unsigned long gPreviousL1 = 0; //   300ms interval  Adress setup mode LED Blink Cicle
unsigned long gPreviousL2 = 0; //   0.5ms　Wait
unsigned long gPreviousL3 = 0; //   Sensor interval
unsigned long gPreviousRecvRailComCutOut = 0;




/*********************************************************************************
 SETUP
 ********************************************************************************/
void setup() {

  /*First initialize the LocoNet interface*/ 
      LocoNet.init(7);

  //LED初期化
      pinMode(DAT_PIN0 , OUTPUT);  // OUTPUTPin1
 //     pinMode(DAT_PIN1 , OUTPUT);  // OUTPUTPin2
 //     pinMode(DAT_PIN2 , OUTPUT);  // OUTPUTPin3
 //     pinMode(DAT_PIN3 , OUTPUT);  // OUTPUTPin4
 //     pinMode(DAT_PIN4 , OUTPUT);  // OUTPUTPin5
 //     pinMode(DAT_PIN5 , OUTPUT);  // OUTPUTPin6
 //     pinMode(DAT_PIN6 , OUTPUT);  // OUTPUTPin7
      pinMode(DAT_PIN7 , INPUT_PULLUP);  // INPUTPin8
      pinMode(SETUP_PIN , INPUT_PULLUP);  // Setup Pin
      pinMode(STAT_LED_PIN , OUTPUT);  // Power LED
      digitalWrite(STAT_LED_PIN, HIGH);
      digitalWrite(DAT_PIN0, LOW);  

 
  //TML637初期化
  tm1637.init();
  tm1637.set(BRIGHT_TYPICAL); //BRIGHT_TYPICAL = 2,BRIGHT_DARKEST = 0,BRIGHTEST = 7;
  tm1637.displayStr("----"); 



  //Reset task
      gPreviousL1 = millis();
      gPreviousL2 = millis();
      gPreviousL3 = millis();
 

  //シリアル通信開始
    Serial.begin(250000);
#ifdef DEBUG
        Serial.println("!Start!");
#endif
    for (int i = 0; i < 3; i++) {
        gRecvDataGram[i].mID = 0;
        gRecvDataGram[i].mData1 = 0;
        gRecvDataGram[i].mData2 = 0;
        gRecvDataGram[i].mData3 = 0;
        gRecvDataGram[i].mData4 = 0;
    }
      
}
 
/*********************************************************************************
  Loop処理
 ********************************************************************************/

void loop() {
 
  /*********************************************
  * Loconet受信処理
  *********************************************/
  // Check for any received LocoNet packets

  LnPacket = LocoNet.receive() ;
  if (LnPacket)  {
      //Serial.println("LnPacket");
      if (!LocoNet.processSwitchSensorMessage(LnPacket)){
        digitalWrite(STAT_LED_PIN, LOW);
        delay(10);
        digitalWrite(STAT_LED_PIN, HIGH);
      }
  }
  
  /*********************************************
  * Bi-Di受信処理
  *********************************************/
    //カットアウト信号がlowになったらinterruptflagを１にしてシリアルポートから受信
    if (((PIND & _BV(2)) == LOW) && (interruptflag == 0)) {
        interruptflag = 1;

        //受信ウェイト(RailComの信号が溜まるまで待つ、通常は0.5msで終了。)
        gWaitTime = TIME_RECVWAIT;

#ifdef DEBUG
        Serial.println("!CUTOUT!");
#endif
      }

      //Railcom Decode
      if (gWaitTime == 0) {
          if (interruptflag == 1) {
            // シリアルポートにデータがいるかチェック
            if (Serial.available()) {

#ifdef DEBUG
                Serial.println("!RECV!");
#endif
                receiveSerial();
            }
            else {
                //CutOUTがDisableになった場合に2秒で表示をリセットします。
                if (no_recive == 0) {
                    no_signaltime = millis();
                    no_recive = 1;
                }
                else if ((millis() - no_signaltime) >= 2000) {
                    command_count = 0;
                    interruptflag = 0;
                    no_recive = 0;
                    clearAddr();

                    //Display Clear
                    tm1637.displayStr("----");
                    if ( gAddress_mem == 1 ){
                      gAddress_mem = 0;
                      sendLnet_command(1, High_add , Low_add, gBoard, gZone, gSection_no, 0);  
                    }
                }
            }
        }
      }

    //Display
   if ((millis() - gPreviousL2) >= 2) {
       //Reset task
        gPreviousL2 = millis();
       //WaitTime
        if (gWaitTime > 0) {
            gWaitTime--;
        }
   
   }

    //在線センサー検出処理
    if ((millis() - gPreviousL3) >= 1000) {
      //Reset task
      gPreviousL3 = millis();
      if (digitalRead(DAT_PIN7) == 0 ){
          if (gSection_mem == 0) {
            //在線検出の場合
            gSection_mem = 1;
            sendLnet_command(0, 0, 0, gBoard, 0, gSection_no, 1);  
            digitalWrite(DAT_PIN0, HIGH);
#ifdef DEBUG
        Serial.println("!Sensor ON!");
#endif
          }
      }
      else {
           //非検出の場合 
          if  (gSection_mem == 1) {
            gSection_mem = 0 ;
            sendLnet_command(0, 0, 0, gBoard, 0, gSection_no, 0);
            digitalWrite(DAT_PIN0, LOW);  
#ifdef DEBUG
            Serial.println("!Sensor OFF!");
#endif
          }
      }
    }

  //変数クリア
  gPresent = 0;
}


/*********************************************************************************
 * Bi-Diシリアル一括受信処理
 * 
 * @param なし
 * @return なし
 *********************************************************************************/

void receiveSerial()
{
    byte aAcks = 0;
    no_recive = 0;

    while (Serial.available() > 0) {
        //データ受信
        receivedByte = Serial.read();
        //データが00hは読み取りません 
        if (receivedByte != 0) {
            gReciveData[command_count] = receivedByte;
            command_count = command_count + 1;
            no_command = 0;
        }
        //データが00hで2秒連続した場合は表示をリセットします。
        else if (receivedByte == 0) {
            if (no_command == 0) {
                no_commandtime = millis();
                no_command = 1;
            }
            else if ((millis() - no_commandtime) >= 2000) {
                command_count = 0;
                interruptflag = 0;
                no_command = 0;
                clearAddr();

                //Display Clear
                tm1637.displayStr("----");
                if ( gAddress_mem == 1 ){
                   gAddress_mem = 0;
                   sendLnet_command(1, High_add , Low_add, gBoard, gZone, gSection_no, 0) ; 
                   return;
                }
            }
        }
    }

 //   Serial.flush();

    if (command_count > 8) {
        command_count = 8;
    }

    //所定のバイト数受信したら受信データを処理します
    //アドレスは2バイトのデータを2回に分けて送信してくるので4個受信します
    if (command_count >= 4) {

#ifdef DEBUG
        Serial.print(" ReceivedByte ");
        for (int i = 0; i < command_count; i++) {
            Serial.print(gReciveData[i], HEX);
            Serial.print(" ");
        }
        Serial.println("");
#endif

        //受信したコマンドを変換テーブルで変換
        excec_conversion(command_count);

        //受信データクリア
        for (int i = 0; i < SIZE_RECV; i++) {
            gReciveData[i] = 0;
        }
 
    }
    //フラグ変数クリア　
    command_count = 0;
    interruptflag = 0;
}

/*********************************************************************************
 * 受信データパケット復号・変換処理
 * 
 * @param [in] 受信サイズ(バイト)
 * @return なし
 *********************************************************************************/

void excec_conversion(byte inRecvSize)
{

    byte aErrChk = 0;
    byte aDblSendChk = 0;
    byte aLen = (inRecvSize / 2);
    byte aJump = 0;

    //受信したデータを変換テーブルで変換
    for (int i = 0; i < inRecvSize; i++) {
        gConversionData[i] = convtable[gReciveData[i]];

        //4/8 codingに合っていない形式の場合はエラーなので、フラグを立てる
        if (gConversionData[i] == 255) {

            switch (gReciveData[i]) {
            case 0x0F:
                //NACK
                gConversionData[i] = 0x00;
                break;
            case 0xF0:
                //ACK
                gConversionData[i] = 0x00;
                break;
            case 0xE1:
                //BUSY
                gConversionData[i] = 0x00;
                break;
            default:
              
              //Kuehn対策。8Fが入り込む。. Ch.1でその後は無視させる
              if( (aErrChk == 0) && (gReciveData[i] == 0x8F) && (i == 2))
              {
                aLen = 1;
                gConversionData[i] = 0x00;
                aJump = 1;
              }
              else
              {
                  aErrChk++;
              }
                break;
            }
        }
      if( aJump == 1)
      {
        break;
      }
    }



    if (aErrChk == 0) {

        //変換した値を下位12ビット分で　1番目と2番目　3番目と4番目で結合
        // 1番目00010110 2番目 00101100 -> 010110 + 101100 -> 010110101100
        //work変数のクリア
        bitwork1 = B00000000;
        bitwork2 = B00000000;
        bitworkA = B00000000;
        bitworkB = B00000000;

        //結合した値を上位4ビットと下位4ビットに分割
        //2桁アドレス
        //1番目と2番目0010.00000011  <-下位8ビットがCV03(2桁アドレス）
        //3番目と4番目0001.00000000  <-上位4ビットが0001で下位8ビットが0であれば2桁アドレス　
        //4桁アドレス
        //1番目と2番目0001.10001110  <-上位4ビットが0001で下位8ビットがCV17
        //3番目と4番目0010.11110110  <-上位4ビットが0010で下位8ビットがCV18
        for (int i = 0; i < aLen; i++) {

            if (gConversionData[(i * 2)] != 0) {
                bitwork1 = gConversionData[(i * 2)];
                bitwork2 = gConversionData[(i * 2) + 1];

                //bitwork1　下位2BitをbitworkAの上位2ビットに移動　（ローテート命令があるといいのに）
                bitWrite(bitworkA, 0, bitRead(bitwork1, 0));
                bitWrite(bitworkA, 1, bitRead(bitwork1, 1));
                bitworkA = bitworkA << 6;
                //上位2ビットでマスクして他のビットをクリア
                bitworkA = bitworkA & B11000000;
                // bitwork2の下位6ビットをマスクし、bitworkA,Bを合成（下位バイトのデータ）
                bitworkB = bitwork2 & B00111111;
                bitworkB = bitworkB |= bitworkA; //下の値(アドレス値）
                //bitwork1の値を下位方向に2ビットシフトして下位4ビットでマスク（上位バイトのデータ）
                bitworkA = bitwork1 >> 2;
                bitworkA = bitworkA & B00001111; //上の値（ID)

                //一旦配列に保持する
                gRecvDataGram[i].mID = bitworkA;
                gRecvDataGram[i].mData1 = bitworkB;

                //IDごとに処理する
                switch (((i > 1) ? 100 : 0) + gRecvDataGram[i].mID) {
                case 0:
                    //POM
                    break;
                case 2:
                    //low addr

                    if ((gAddrReceived & CHK_RCID2) > 0) {
                        //同じIDが複数回来ても一致していればOK、違うならパケット異常

                        if (gAddrDataLow != gRecvDataGram[i].mData1) {
                            clearAddr();

//デバッグ用
#ifdef DEBUG
                            Serial.println("addr chk err!");
#endif
                        }
                    }
                    else {
                        gAddrDataLow = gRecvDataGram[i].mData1;
                        gAddrReceived |= CHK_RCID2;
                    }

                    break;
                case 1:
                    //high addr

                    if ((gAddrReceived & CHK_RCID1) > 0) {
                        //同じIDが複数回来ても一致していればOK、違うならパケット異常

                        if (gAddrDataHigh != gRecvDataGram[i].mData1) {
                            clearAddr();
//デバッグ用
#ifdef DEBUG
                            Serial.println("addr chk err!");
#endif
                        }
                    }
                    else {
                        gAddrDataHigh = gRecvDataGram[i].mData1;
                        gAddrReceived |= CHK_RCID1;
                    }

                    break;

                case 100:
                    //POM
                    break;
                case 103:
                    //ext(MOB)
                    break;
                case 104:
                    //stat1(STAT)
                    break;
                case 105:
                    //time(STAT)
                    break;
                case 106:
                    //err(STAT)
                    break;
                case 007:
                    break;
                case 107:
                    //dyn Speed(MOB,STAT)
                    break;
                case 108:
                    //stat2(STAT)
                    break;
                case 112:
                    //subID(MOB,STAT)
                    break;

                default:
                    //その他のID
                    break;
                }

//デバッグ用
#ifdef DEBUG
                Serial.print("id=");
                Serial.print(gRecvDataGram[i].mID);
                Serial.println("");
                Serial.print("Data=");
                Serial.println(gRecvDataGram[i].mData1);
#endif
            }
        }

        //アドレスの取得
        if (gAddrReceived == 3) {
            if ((gAddrDataLow != 0) && (gAddrDataHigh != 0)) {
                //4digit
                gDetectedLocAddress = ((gAddrDataHigh & 63) * 256) + gAddrDataLow;
            }
            else if (gAddrDataLow != 0) {
                //2digit
                gDetectedLocAddress = gAddrDataLow;
            }
            else {
                //2digit
                gDetectedLocAddress = 0;
            }
        }
        else {
            gDetectedLocAddress = 0;
        }
//デバッグ用
#ifdef DEBUG
        Serial.print(" LocoAddress ");
        Serial.println(gDetectedLocAddress);
#endif

    //　Lnetアドレスコマンド送信用にHighアドレスとLowアドレス設定
        Low_add = gDetectedLocAddress  & 127;
        High_add = gDetectedLocAddress >> 7;
        High_add = High_add & 127;

        if (gDetectedLocAddress > 0) {
            //LEDに数字表示
            tm1637.displayNum((float)gDetectedLocAddress);
            if (gAddress_mem == 0 ){
               gAddress_mem = 1;
               sendLnet_command(1, High_add , Low_add, gBoard, gZone, gSection_no, 1);  
            }
        }
    }
    else {
//デバッグ用
#ifdef DEBUG
        Serial.println("Packet err!");
#endif
    }
}

/*********************************************************************************
 * 保持アドレス関連クリア処理
 * 
 * @param なし
 * @return なし
 *********************************************************************************/

void clearAddr()
{
    gAddrDataLow = 0;
    gAddrDataHigh = 0;
    gAddrReceived = 0;

    //Clear
    for (int i = 0; i < SIZE_CONV; i++) {
        gRecvDataGram[i].mID = 0;
        gRecvDataGram[i].mData1 = 0;
    }
}

/***********************************************************************
  各種関数
************************************************************************/
//以下はLocoNetでのコールバック関数
//--------------------------------------------------------------------------------------------------
//Notifications for accessory control requests
//--------------------------------------------------------------------------------------------------
void notifySwitchRequest( uint16_t Address, uint8_t outData, uint8_t dirData) {
//  Serial.print("notifySwitchRequest():");
//  Serial.print(Address,DEC);
//  Serial.print(",");
//  Serial.print(outData,DEC); 
//  Serial.print(",");
//  Serial.print(dirData,DEC); // 0:t:分岐  32:c:直線
//  Serial.println("");
}
//--------------------------------------------------------------------------------------------------
// Transponding and Multi-Sense Device info Call-back functions
//--------------------------------------------------------------------------------------------------
void notifyMultiSenseTransponder( uint16_t Address, uint8_t Zone, uint16_t LocoAddress, uint8_t Present ){
//    gBoard = ((Address / 16) + 1) ;
//    gPresent = Present;
//    gZone = Zone;
//    gLocoAddress = LocoAddress;
//    int section = (Address - (( Address / 16) * 16));
//    char WK_Zone[] = {Zone};
 //   Serial.print("notifyMultiSenseTransponder: ");
 //   Serial.print("BoardID: ");
 //   Serial.print(gBoard);
 //   Serial.print(", Section: ");
 //   Serial.print(section );
 //   Serial.print(", Zone:");   
 //   Serial.print(Zone); 
 //   Serial.print(", WKZone:");
 //   Serial.print(WK_Zone);  // ZONE  65->A 66->B 67->C 68->: 69->E 70->F 71->G 72->H
 //   Serial.print(", LocoAddress:");
 //   Serial.print(LocoAddress,DEC); 
 //   Serial.print(", Present:");
 //   Serial.print(Present,DEC); 
 //   Serial.println("");
}

//--------------------------------------------------------------------------------------------------


//以下はLocoNetでのコマンド送信関数
/***********************************************
 * Lnetコマンド送信
 * 
 * @param 　state ,Haddress ,Laddress
 * @return なし
 ***********************************************/
void sendLnet_command(uint8_t ptr_code, uint8_t Haddress, uint8_t Laddress, uint16_t gBoard_Add, uint8_t Zone, uint8_t Section_no, uint8_t state) 
{

  if (ptr_code == 0){  //０は在線
      //セクション在線onコマンド送信
      Send_Section_Command(gBoard_Add, Section_no, state);
  }
  else{
      //アドレス コマンド送信 Absent（検出）
      Send_Detect_Command(gBoard_Add, Zone, Haddress, Laddress, state);
  }
}

void Send_Detect_Command(uint16_t BD_Add, uint8_t Zone, uint8_t UpperLocoAddress, uint8_t LowerLocoAddress, uint8_t Present) {
/********************************************************************************************** 
 *  //Send_Adress_Command//
 *  LoconetのBDL168相当のTransPondingのコマンド
 *  先頭が0xD0で始まる6バイトのコマンド
 *  D0 aa bb cc dd ee
 *  aa 　=　0 0 n A7 A6 A5 A4 A3
 *  bb 　=　0 A2 A1 A0 Z3 Z2 Z1 Z0
 *  Present=1 Absent（検出） Present=0 Present（未検出）
 *  A7～A0　BDL168のアドレス
 *  Z3～Z0　ゾーン番号
 *  cc 　=　下位7ビットが車輌アドレスの上位アドレス
 *  　　　（2桁アドレスの場合は0x7D）
 *  dd 　=　下位7ビットが車輌アドレスの下位アドレス
 *          上位アドレス+下位アドレスの並びでビットから１０進数に変換する
 *  　　　  と車輌アドレスになる
 *   　　　 2桁アドレスの場合は　下位アドレスのみ１０進数に変換
 *  ee 　=　check SUM (チェックサムを含めXORしてFFになる値）
************************************************************************************************/
   sendwork_aa = 0 ;
   sendwork_bb = 0 ;
    
   //ボードアドレスはマイナス１する　0がボードアドレス１となる 
   BD_Add = BD_Add - 1 ;
      
   lnMsg SendPacket ;
   //ボードアドレスの下位3ビットをｂｂのbit4-6にセット　Bit7に0をセット
   bitWrite(sendwork_bb, 7, 0 );
   bitWrite(sendwork_bb, 6, bitRead(BD_Add, 2));
   bitWrite(sendwork_bb, 5, bitRead(BD_Add, 1));
   bitWrite(sendwork_bb, 4, bitRead(BD_Add, 0));
   //bbの下位４ビットにORでゾーン番号をマスク
   Zone = Zone &  B00001111;
   sendwork_bb = sendwork_bb |= Zone ; 
   //aaにボードアドレスをセットして右に3ビットローテ
   sendwork_aa = BD_Add >> 3;
   sendwork_aa = sendwork_aa & B00011111 ; 
   //aaの上位2ビットbit6-7に0設定　検出フラグをbit5にセット
   bitWrite(sendwork_aa, 7, 0 );
   bitWrite(sendwork_aa, 6, 0 );
   bitWrite(sendwork_aa, 5, bitRead(Present, 0));
   //cc ddのアドレスセット
   if (UpperLocoAddress == 0) {
      sendwork_cc = 0x7D;   //2桁アドレスの対応
   } else {
      sendwork_cc = UpperLocoAddress ;
   }
   sendwork_dd = LowerLocoAddress ;
   //チェックサム計算
   sendchksum = 0xD0 ;
   sendchksum = sendchksum ^ sendwork_aa ;
   sendchksum = sendchksum ^ sendwork_bb ;
   sendchksum = sendchksum ^ sendwork_cc ;
   sendchksum = sendchksum ^ sendwork_dd ;
   sendchksum = ~sendchksum ;
   //データのセット 
   SendPacket.data[ 0 ] = 0xD0 ;
   SendPacket.data[ 1 ] = sendwork_aa ;
   SendPacket.data[ 2 ] = sendwork_bb ;
   SendPacket.data[ 3 ] = sendwork_cc ;
   SendPacket.data[ 4 ] = sendwork_dd ;
   SendPacket.data[ 5 ] = sendchksum ;
   LocoNet.send( &SendPacket );
}

void Send_Section_Command(uint16_t BD_Add, uint8_t Section, uint8_t Present) {
/***********************************************************************************  
 * /// Send_Detect_Command///
 *  LoconetのBDL168からくるSection　On/Off のコマンド
 *  先頭が0xB2で始まる4バイトのコマンド　　　　
 *  B2 aa bb cc
 *     aa  =　0 A3 A2 A1　 A0 S3 S2 S1
 *     bb  =　0  n S0  L   A7 A6 A5 A4
 *            n=1 常時１   L=1 検出 　L=0 未検出
 *            A7～A0　BDL168のアドレス
 *            S3～S0　セクション番号（0~15)
 *     cc　=　check SUM;
************************************************************************************/  
   sendwork_aa = 0 ;
   sendwork_bb = 0 ;
   //ボードアドレスはマイナス１する　0がボードアドレス１となる 
   BD_Add = BD_Add - 1 ;
   //Section1番号もマイナス１する　0がセクション１となる 
   Section = Section - 1 ;
   lnMsg SendPacket ;
   //変数aaへのセット
   bitWrite(sendwork_aa, 7, 0 );
   bitWrite(sendwork_aa, 6, bitRead(BD_Add, 3));
   bitWrite(sendwork_aa, 5, bitRead(BD_Add, 2));
   bitWrite(sendwork_aa, 4, bitRead(BD_Add, 1));
   bitWrite(sendwork_aa, 3, bitRead(BD_Add, 0));
   bitWrite(sendwork_aa, 2, bitRead(Section, 3));
   bitWrite(sendwork_aa, 1, bitRead(Section, 2));
   bitWrite(sendwork_aa, 0, bitRead(Section, 1));
   //変数bbのセット
   bitWrite(sendwork_bb, 7, 0 );
   bitWrite(sendwork_bb, 6, 1 );
   bitWrite(sendwork_bb, 5, bitRead(Section, 0));
   bitWrite(sendwork_bb, 4, bitRead(Present, 0));
   bitWrite(sendwork_bb, 3, bitRead(BD_Add, 7));
   bitWrite(sendwork_bb, 2, bitRead(BD_Add, 6));
   bitWrite(sendwork_bb, 1, bitRead(BD_Add, 5));
   bitWrite(sendwork_bb, 0, bitRead(BD_Add, 4));
   //チェックサム計算
   sendchksum = 0xB2;
   sendchksum = sendchksum ^ sendwork_aa ;
   sendchksum = sendchksum ^ sendwork_bb ;
   sendchksum = ~sendchksum ;
   //データのセット 
   SendPacket.data[ 0 ] = 0xB2 ;
   SendPacket.data[ 1 ] = sendwork_aa ;
   SendPacket.data[ 2 ] = sendwork_bb ;
   SendPacket.data[ 3 ] = sendchksum ;
   LocoNet.send( &SendPacket );
}
